import json
import os

import colander
from pyramid.security import NO_PERMISSION_REQUIRED

from kinto.core import Service


HERE = os.path.dirname(__file__)
ORIGIN = os.path.dirname(HERE)  # package root.
_CONTRIBUTE_INFO = None

contribute = Service(
    name="contribute.json", description="Open-source information", path="/contribute.json"
)


class ContributeResponseSchema(colander.MappingSchema):
    body = colander.SchemaNode(colander.Mapping(unknown="preserve"))


contribute_responses = {
    "200": ContributeResponseSchema(description="Return open source contributing information.")
}


@contribute.get(
    permission=NO_PERMISSION_REQUIRED,
    tags=["Utilities"],
    operation_id="contribute",
    response_schemas=contribute_responses,
)
def contribute_get(request):
    global _CONTRIBUTE_INFO
    if _CONTRIBUTE_INFO is None:
        with open(os.path.join(ORIGIN, "contribute.json")) as f:
            _CONTRIBUTE_INFO = json.load(f)
    return _CONTRIBUTE_INFO



import colander

from kinto.core.cornice.validators import colander_body_validator


def trim(docstring):
    """
    Remove the tabs to spaces, and remove the extra spaces / tabs that are in
    front of the text in docstrings.

    Implementation taken from http://www.python.org/dev/peps/pep-0257/
    """
    if not docstring:
        return ""
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    lines = [line.strip() for line in lines]
    res = "\n".join(lines)
    return res


def body_schema_transformer(schema, args):
    validators = args.get("validators", [])
    if colander_body_validator in validators:
        body_schema = schema
        schema = colander.MappingSchema()
        schema["body"] = body_schema
    return schema


def merge_dicts(base, changes):
    """Merge b into a recursively, without overwriting values.

    :param base: the dict that will be altered.
    :param changes: changes to update base.
    """
    for k, v in changes.items():
        if isinstance(v, dict):
            merge_dicts(base.setdefault(k, {}), v)
        else:
            base.setdefault(k, v)



import logging
import os
import shutil
import warnings
from time import perf_counter as time_now

from pyramid.exceptions import ConfigurationError
from pyramid.response import Response
from pyramid.settings import asbool
from zope.interface import implementer

from kinto.core import metrics
from kinto.core.utils import safe_wraps


try:
    import prometheus_client as prometheus_module
except ImportError:  # pragma: no cover
    prometheus_module = None


logger = logging.getLogger(__name__)

_METRICS = {}
_REGISTRY = None


PROMETHEUS_MULTIPROC_DIR = os.getenv("PROMETHEUS_MULTIPROC_DIR")


def get_registry():
    global _REGISTRY

    if _REGISTRY is None:
        if PROMETHEUS_MULTIPROC_DIR:  # pragma: no cover
            from prometheus_client import multiprocess

            _reset_multiproc_folder_content()
            # Ref: https://prometheus.github.io/client_python/multiprocess/
            _REGISTRY = prometheus_module.CollectorRegistry()
            multiprocess.MultiProcessCollector(_REGISTRY)
        else:
            _REGISTRY = prometheus_module.REGISTRY
            logger.warning("Prometheus metrics will run in single-process mode only.")
    return _REGISTRY


def _fix_metric_name(s):
    return s.replace("-", "_").replace(".", "_").replace(" ", "_")


class Timer:
    """
    A decorator to time the execution of a function. It will use the
    `prometheus_client.Histogram` to record the time taken by the function
    in seconds. The histogram is passed as an argument to the
    constructor.

    Main limitation: it does not support `labels` on the decorator.
    """

    def __init__(self, histogram):
        self.histogram = histogram
        self._start_time = None

    def set_labels(self, labels):
        if not labels:
            return
        self.histogram = self.histogram.labels(*(label_value for _, label_value in labels))

    def observe(self, value):
        return self.histogram.observe(value)

    def __call__(self, f):
        @safe_wraps(f)
        def _wrapped(*args, **kwargs):
            start_time = time_now()
            try:
                return f(*args, **kwargs)
            finally:
                dt_sec = time_now() - start_time
                self.histogram.observe(dt_sec)

        return _wrapped

    def __enter__(self):
        return self.start()

    def __exit__(self, typ, value, tb):
        self.stop()

    def start(self):
        self._start_time = time_now()
        return self

    def stop(self):
        if self._start_time is None:  # pragma: nocover
            raise RuntimeError("Timer has not started.")
        dt_sec = time_now() - self._start_time
        self.histogram.observe(dt_sec)
        return self


@implementer(metrics.IMetricsService)
class PrometheusService:
    def __init__(self, prefix=""):
        prefix_clean = ""
        if prefix:
            # In GCP Console, the metrics are grouped by the first
            # word before the first underscore. Here we make sure the specified
            # prefix is not mixed up with metrics names.
            # (eg. `remote-settings` -> `remotesettings_`, `kinto_` -> `kinto_`)
            prefix_clean = _fix_metric_name(prefix).replace("_", "") + "_"
        self.prefix = prefix_clean.lower()

    def timer(self, key, value=None, labels=[]):
        global _METRICS
        key = self.prefix + key

        if key not in _METRICS:
            _METRICS[key] = prometheus_module.Histogram(
                _fix_metric_name(key),
                f"Histogram of {key}",
                registry=get_registry(),
                labelnames=[label_name for label_name, _ in labels],
            )

        if not isinstance(_METRICS[key], prometheus_module.Histogram):
            raise RuntimeError(
                f"Metric {key} already exists with different type ({_METRICS[key]})"
            )

        timer = Timer(_METRICS[key])
        timer.set_labels(labels)

        if value is not None:
            # We are timing something.
            return timer.observe(value)

        # We are not timing anything, just returning the timer object
        # (eg. to be used as decorator or context manager).
        # Note that in this case, the labels values will be the same for all calls.
        return timer

    def observe(self, key, value, labels=[]):
        global _METRICS
        key = self.prefix + key

        if key not in _METRICS:
            _METRICS[key] = prometheus_module.Summary(
                _fix_metric_name(key),
                f"Summary of {key}",
                labelnames=[label_name for label_name, _ in labels],
                registry=get_registry(),
            )

        if not isinstance(_METRICS[key], prometheus_module.Summary):
            raise RuntimeError(
                f"Metric {key} already exists with different type ({_METRICS[key]})"
            )

        m = _METRICS[key]
        if labels:
            m = m.labels(*(label_value for _, label_value in labels))

        m.observe(value)

    def count(self, key, count=1, unique=None):
        global _METRICS
        key = self.prefix + key

        labels = []

        if unique:
            if isinstance(unique, str):
                warnings.warn(
                    "`unique` parameter should be of type ``list[tuple[str, str]]``",
                    DeprecationWarning,
                )
                # Turn `unique` into a group and a value:
                # "bob" -> "group.bob"
                # "method.basicauth.mat" -> [("method_basicauth", "mat")]`
                if "." not in unique:
                    unique = f"group.{unique}"
                label_name, label_value = unique.rsplit(".", 1)
                unique = [(label_name, label_value)]

            labels = [
                (_fix_metric_name(label_name), label_value) for label_name, label_value in unique
            ]

        if key not in _METRICS:
            _METRICS[key] = prometheus_module.Counter(
                _fix_metric_name(key),
                f"Counter of {key}",
                labelnames=[label_name for label_name, _ in labels],
                registry=get_registry(),
            )

        if not isinstance(_METRICS[key], prometheus_module.Counter):
            raise RuntimeError(
                f"Metric {key} already exists with different type ({_METRICS[key]})"
            )

        m = _METRICS[key]
        if labels:
            m = m.labels(*(label_value for _, label_value in labels))

        m.inc(count)


def metrics_view(request):
    registry = get_registry()
    data = prometheus_module.generate_latest(registry)
    resp = Response(body=data)
    resp.headers["Content-Type"] = prometheus_module.CONTENT_TYPE_LATEST
    resp.headers["Content-Length"] = str(len(data))
    return resp


def _reset_multiproc_folder_content():  # pragma: no cover
    shutil.rmtree(PROMETHEUS_MULTIPROC_DIR, ignore_errors=True)
    os.makedirs(PROMETHEUS_MULTIPROC_DIR, exist_ok=True)


def includeme(config):
    if prometheus_module is None:
        error_msg = (
            "Please install Kinto with monitoring dependencies (e.g. prometheus-client package)"
        )
        raise ConfigurationError(error_msg)

    settings = config.get_settings()

    if not asbool(settings.get("prometheus_created_metrics_enabled", True)):
        prometheus_module.disable_created_metrics()

    config.add_api_capability(
        "prometheus",
        description="Prometheus metrics.",
        url="https://github.com/Kinto/kinto/",
    )

    config.add_route("prometheus_metrics", "/__metrics__")
    config.add_view(metrics_view, route_name="prometheus_metrics")

    # Reinitialize the registry on initialization.
    # This is mainly useful in tests, where the plugin is included
    # several times with different settings.
    registry = get_registry()

    for collector in _METRICS.values():
        try:
            registry.unregister(collector)
        except KeyError:  # pragma: no cover
            pass
    _METRICS.clear()

    prefix = settings.get("prometheus_prefix", settings["project_name"])

    config.registry.registerUtility(PrometheusService(prefix=prefix), metrics.IMetricsService)



import unittest
from unittest import mock

from pyramid.config import Configurator

from kinto.core import metrics


class TestedClass:
    attribute = 3.14

    def test_method(self):
        pass

    def _private_method(self):
        pass


class WatchExecutionTimeTest(unittest.TestCase):
    def setUp(self):
        self.test_object = TestedClass()
        self.mocked = mock.MagicMock()
        metrics.watch_execution_time(self.mocked, self.test_object, prefix="test")

    def test_public_methods_generates_statsd_calls(self):
        self.test_object.test_method()
        self.mocked.timer.assert_called_with(
            "test.testedclass.seconds", labels=[("method", "test_method")]
        )

    def test_private_methods_does_not_generates_statsd_calls(self):
        self.test_object._private_method()
        self.assertFalse(self.mocked().timer.called)


class ListenerWithTimerTest(unittest.TestCase):
    def setUp(self):
        self.config = Configurator()
        self.func = lambda x: x  # noqa: E731

    def test_without_metrics_service(self):
        wrapped = metrics.listener_with_timer(self.config, "key", self.func)

        self.assertEqual(wrapped(42), 42)  # does not raise

    def test_with_metrics_service(self):
        self.config.registry.registerUtility(mock.MagicMock(), metrics.IMetricsService)
        wrapped = metrics.listener_with_timer(self.config, "key", self.func)

        self.assertEqual(wrapped(42), 42)
        self.config.registry.metrics.timer.assert_called_with("key.seconds")



from pyramid.security import NO_PERMISSION_REQUIRED

from kinto.core.cornice_swagger.swagger import CorniceSwagger


__author__ = """Josip Delic"""
__email__ = "delicj@delijati.net"
__version__ = "0.3.0"


__all__ = ["CorniceSwagger"]


class CorniceSwaggerPredicate(object):
    """Predicate to add simple information to Cornice Swagger."""

    def __init__(self, schema, config):
        self.schema = schema

    def phash(self):
        return str(self.schema)

    def __call__(self, context, request):
        return self.schema


def includeme(config):
    # Custom view parameters
    config.add_view_predicate("response_schemas", CorniceSwaggerPredicate)
    config.add_view_predicate("tags", CorniceSwaggerPredicate)
    config.add_view_predicate("operation_id", CorniceSwaggerPredicate)
    config.add_view_predicate("api_security", CorniceSwaggerPredicate)
    config.add_directive("cornice_enable_openapi_view", cornice_enable_openapi_view)
    config.add_directive("cornice_enable_openapi_explorer", cornice_enable_openapi_explorer)


def cornice_enable_openapi_view(
    config,
    api_path="/api-explorer/swagger.json",
    permission=NO_PERMISSION_REQUIRED,
    route_factory=None,
    **kwargs,
):
    """
    :param config:
        Pyramid configurator object
    :param api_path:
        where to expose swagger JSON definition view
    :param permission:
        pyramid permission for those views
    :param route_factory:
        factory for context object for those routes
    :param kwargs:
        kwargs that will be passed to CorniceSwagger's `generate()`

    This registers and configures the view that serves api definitions
    """
    config.registry.settings["cornice_swagger.spec_kwargs"] = kwargs
    config.add_route("cornice_swagger.open_api_path", api_path, factory=route_factory)
    config.add_view(
        "cornice_swagger.views.open_api_json_view",
        renderer="json",
        permission=permission,
        route_name="cornice_swagger.open_api_path",
    )


def cornice_enable_openapi_explorer(
    config,
    api_explorer_path="/api-explorer",
    permission=NO_PERMISSION_REQUIRED,
    route_factory=None,
    **kwargs,
):
    """
    :param config:
        Pyramid configurator object
    :param api_explorer_path:
        where to expose Swagger UI interface view
    :param permission:
        pyramid permission for those views
    :param route_factory:
        factory for context object for those routes

    This registers and configures the view that serves api explorer
    """
    config.add_route("cornice_swagger.api_explorer_path", api_explorer_path, factory=route_factory)
    config.add_view(
        "cornice_swagger.views.swagger_ui_template_view",
        permission=permission,
        route_name="cornice_swagger.api_explorer_path",
    )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import inspect


def _generate_marshmallow_validator(location):
    """
    Generate a marshmallow validator for data from the given location.

    :param location: The location in the request to find the data to be
        validated, such as "body" or "querystring".
    :type location: str
    :return: Returns a callable that will validate the request at the given
        location.
    :rtype: callable
    """

    def _validator(request, schema=None, deserializer=None, **kwargs):
        """
        Validate the location against the schema defined on the service.

        The content of the location is deserialized, validated and stored in
        the ``request.validated`` attribute.

        Keyword arguments to be included when initialising the marshmallow
        schema can be passed as a dict in ``kwargs['schema_kwargs']`` variable.

        .. note::

            If no schema is defined, this validator does nothing.

        :param request: Current request
        :type request: :class:`~pyramid:pyramid.request.Request`

        :param schema: The marshmallow schema
        :param deserializer: Optional deserializer, defaults to
            :func:`cornice.validators.extract_cstruct`
        """
        import marshmallow
        import marshmallow.schema
        from marshmallow.utils import EXCLUDE

        if schema is None:
            return

        # see if the user wants to set any keyword arguments for their schema
        schema_kwargs = kwargs.get("schema_kwargs", {})
        schema = _instantiate_schema(schema, **schema_kwargs)

        class ValidatedField(marshmallow.fields.Field):
            def _deserialize(self, value, attr, data, **kwargs):
                schema.context.setdefault("request", request)
                deserialized = schema.load(value)
                return deserialized

        class Meta(object):
            strict = True
            ordered = True
            unknown = EXCLUDE

        class RequestSchemaMeta(marshmallow.schema.SchemaMeta):
            """
            A metaclass that will inject a location class attribute into
            RequestSchema.
            """

            def __new__(cls, name, bases, class_attrs):
                """
                Instantiate the RequestSchema class.

                :param name: The name of the class we are instantiating. Will
                    be "RequestSchema".
                :type name: str
                :param bases: The class's superclasses.
                :type bases: tuple
                :param dct: The class's class attributes.
                :type dct: dict
                """

                class_attrs[location] = ValidatedField(
                    required=True, metadata={"load_from": location}
                )
                class_attrs["Meta"] = Meta
                return type(name, bases, class_attrs)

        class RequestSchema(marshmallow.Schema, metaclass=RequestSchemaMeta):  # noqa
            """A schema to validate the request's location attributes."""

            pass

        validator(request, RequestSchema, deserializer, **kwargs)
        request.validated = request.validated.get(location, {})

    return _validator


body_validator = _generate_marshmallow_validator("body")
headers_validator = _generate_marshmallow_validator("header")
path_validator = _generate_marshmallow_validator("path")
querystring_validator = _generate_marshmallow_validator("querystring")


def _message_normalizer(exc, no_field_name="_schema"):
    """
    Normally `normalize_messages` will exist on `ValidationError` but pre 2.10
    versions don't have it
    :param exc:
    :param no_field_name:
    :return:
    """
    if isinstance(exc.messages, dict):
        return exc.messages
    field_names = exc.kwargs.get("field_names", [])
    if len(field_names) == 0:
        return {no_field_name: exc.messages}
    return dict((name, exc.messages) for name in field_names)


def validator(request, schema=None, deserializer=None, **kwargs):
    """
    Validate the full request against the schema defined on the service.

    Each attribute of the request is deserialized, validated and stored in the
    ``request.validated`` attribute
    (eg. body in ``request.validated['body']``).

    .. note::

        If no schema is defined, this validator does nothing.

    :param request: Current request
    :type request: :class:`~pyramid:pyramid.request.Request`

    :param schema: The marshmallow schema
    :param deserializer: Optional deserializer, defaults to
        :func:`cornice.validators.extract_cstruct`
    """
    import marshmallow

    from kinto.core.cornice.validators import extract_cstruct

    if deserializer is None:
        deserializer = extract_cstruct

    if schema is None:
        return

    schema = _instantiate_schema(schema)
    schema.context.setdefault("request", request)

    cstruct = deserializer(request)
    try:
        deserialized = schema.load(cstruct)
    except marshmallow.ValidationError as err:
        # translate = request.localizer.translate
        normalized_errors = _message_normalizer(err)
        for location, details in normalized_errors.items():
            location = location if location != "_schema" else ""
            if hasattr(details, "items"):
                for subfield, msg in details.items():
                    request.errors.add(location, subfield, msg)
            else:
                request.errors.add(location, location, details)
    else:
        request.validated.update(deserialized)


def _instantiate_schema(schema, **kwargs):
    """
    Returns an object of the given marshmallow schema.

    :param schema: The marshmallow schema class with which the request should
        be validated
    :param kwargs: The keyword arguments that will be provided to the
        marshmallow schema's constructor
    :return: The object of the marshmallow schema
    """
    if not inspect.isclass(schema):
        raise ValueError("You need to pass Marshmallow class instead of schema instance")
    return schema(**kwargs)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
import fnmatch
import functools

from pyramid.settings import asbool


CORS_PARAMETERS = (
    "cors_headers",
    "cors_enabled",
    "cors_origins",
    "cors_credentials",
    "cors_max_age",
    "cors_expose_all_headers",
)


def get_cors_preflight_view(service):
    """Return a view for the OPTION method.

    Checks that the User-Agent is authorized to do a request to the server, and
    to this particular service, and add the various checks that are specified
    in http://www.w3.org/TR/cors/#resource-processing-model.
    """

    def _preflight_view(request):
        response = request.response
        origin = request.headers.get("Origin")
        supported_headers = service.cors_supported_headers_for()

        if not origin:
            request.errors.add("header", "Origin", "this header is mandatory")

        requested_method = request.headers.get("Access-Control-Request-Method")
        if not requested_method:
            request.errors.add(
                "header", "Access-Control-Request-Method", "this header is mandatory"
            )

        if not (requested_method and origin):
            return

        requested_headers = request.headers.get("Access-Control-Request-Headers", ())

        if requested_headers:
            requested_headers = map(str.strip, requested_headers.split(","))

        if requested_method not in service.cors_supported_methods:
            request.errors.add("header", "Access-Control-Request-Method", "Method not allowed")

        if not service.cors_expose_all_headers:
            for h in requested_headers:
                if h.lower() not in [s.lower() for s in supported_headers]:
                    request.errors.add(
                        "header", "Access-Control-Request-Headers", 'Header "%s" not allowed' % h
                    )

        supported_headers = set(supported_headers) | set(requested_headers)

        response.headers["Access-Control-Allow-Headers"] = ",".join(supported_headers)

        response.headers["Access-Control-Allow-Methods"] = ",".join(service.cors_supported_methods)

        max_age = service.cors_max_age_for(requested_method)
        if max_age is not None:
            response.headers["Access-Control-Max-Age"] = str(max_age)

        return None

    return _preflight_view


def _get_method(request):
    """Return what's supposed to be the method for CORS operations.
    (e.g if the verb is options, look at the A-C-Request-Method header,
    otherwise return the HTTP verb).
    """
    if request.method == "OPTIONS":
        method = request.headers.get("Access-Control-Request-Method", request.method)
    else:
        method = request.method
    return method


def ensure_origin(service, request, response=None, **kwargs):
    """Ensure that the origin header is set and allowed."""
    response = response or request.response

    # Don't check this twice.
    if not request.info.get("cors_checked", False):
        method = _get_method(request)

        origin = request.headers.get("Origin")

        if not origin:
            always_cors = asbool(request.registry.settings.get("cornice.always_cors"))
            # With this setting, if the service origins has "*", then
            # always return CORS headers.
            origins = getattr(service, "cors_origins", [])
            if always_cors and "*" in origins:
                origin = "*"

        if origin:
            if not any([fnmatch.fnmatchcase(origin, o) for o in service.cors_origins_for(method)]):
                request.errors.add("header", "Origin", "%s not allowed" % origin)
            elif service.cors_support_credentials_for(method):
                response.headers["Access-Control-Allow-Origin"] = origin
            else:
                if any([o == "*" for o in service.cors_origins_for(method)]):
                    response.headers["Access-Control-Allow-Origin"] = "*"
                else:
                    response.headers["Access-Control-Allow-Origin"] = origin
        request.info["cors_checked"] = True
    return response


def get_cors_validator(service):
    return functools.partial(ensure_origin, service)


def apply_cors_post_request(service, request, response):
    """Handles CORS-related post-request things.

    Add some response headers, such as the Expose-Headers and the
    Allow-Credentials ones.
    """
    response = ensure_origin(service, request, response)
    method = _get_method(request)

    if (
        service.cors_support_credentials_for(method)
        and "Access-Control-Allow-Credentials" not in response.headers
    ):
        response.headers["Access-Control-Allow-Credentials"] = "true"

    if request.method != "OPTIONS":
        # Which headers are exposed?
        supported_headers = service.cors_supported_headers_for(request.method)
        if supported_headers:
            response.headers["Access-Control-Expose-Headers"] = ", ".join(supported_headers)

    return response



"""Cornice Swagger 2.0 documentor"""

import inspect
import warnings
from collections import OrderedDict

import colander
from pyramid.threadlocal import get_current_registry

from kinto.core.cornice import Service
from kinto.core.cornice.util import to_list
from kinto.core.cornice_swagger.converters import (
    ParameterConversionDispatcher as ParameterConverter,
)
from kinto.core.cornice_swagger.converters import TypeConversionDispatcher as TypeConverter
from kinto.core.cornice_swagger.util import body_schema_transformer, merge_dicts, trim


class CorniceSwaggerException(Exception):
    """Raised when cornice services have structural problems to be converted."""


class DefinitionHandler(object):
    """Handles Swagger object definitions provided by cornice as colander schemas."""

    json_pointer = "#/definitions/"

    def __init__(self, ref=0, type_converter=TypeConverter()):
        """
        :param ref:
            The depth that should be used by self.ref when calling self.from_schema.
        """

        self.definition_registry = {}
        self.ref = ref
        self.type_converter = type_converter

    def from_schema(self, schema_node, base_name=None):
        """
        Creates a Swagger definition from a colander schema.

        :param schema_node:
            Colander schema to be transformed into a Swagger definition.
        :param base_name:
            Schema alternative title.

        :rtype: dict
        :returns: Swagger schema.
        """
        return self._ref_recursive(self.type_converter(schema_node), self.ref, base_name)

    def _ref_recursive(self, schema, depth, base_name=None):
        """
        Dismantle nested swagger schemas into several definitions using JSON pointers.
        Note: This can be dangerous since definition titles must be unique.

        :param schema:
            Base swagger schema.
        :param depth:
            How many levels of the swagger object schemas should be split into
            swaggger definitions with JSON pointers. Default (0) is no split.
            You may use negative values to split everything.
        :param base_name:
            If schema doesn't have a name, the caller may provide it to be
            used as reference.

        :rtype: dict
        :returns:
            JSON pointer to the root definition schema,
            or the original definition if depth is zero.
        """

        if depth == 0:
            return schema

        if schema["type"] != "object":
            return schema

        name = base_name or schema["title"]

        pointer = self.json_pointer + name
        for child_name, child in schema.get("properties", {}).items():
            schema["properties"][child_name] = self._ref_recursive(child, depth - 1)

        self.definition_registry[name] = schema

        return {"$ref": pointer}


class ParameterHandler(object):
    """Handles swagger parameter definitions."""

    json_pointer = "#/parameters/"

    def __init__(
        self,
        definition_handler=DefinitionHandler(),
        ref=False,
        type_converter=TypeConverter(),
        parameter_converter=ParameterConverter(TypeConverter()),
    ):
        """
        :param definition_handler:
            Callable that handles swagger definition schemas.
        :param ref:
            Specifies the ref value when calling from_xxx methods.
        """

        self.parameter_registry = {}

        self.type_converter = type_converter
        self.parameter_converter = parameter_converter
        self.definitions = definition_handler
        self.ref = ref

    def from_schema(self, schema_node):
        """
        Creates a list of Swagger params from a colander request schema.

        :param schema_node:
            Request schema to be transformed into Swagger.
        :param validators:
            Validators used in colander with the schema.

        :rtype: list
        :returns: List of Swagger parameters.
        """

        params = []

        for param_schema in schema_node.children:
            location = param_schema.name
            if location == "body":
                name = param_schema.__class__.__name__
                if name == "body":
                    name = schema_node.__class__.__name__ + "Body"
                param = self.parameter_converter(location, param_schema)
                param["name"] = name
                if self.ref:
                    param = self._ref(param)
                params.append(param)

            elif location in (("path", "header", "headers", "querystring", "GET")):
                for node_schema in param_schema.children:
                    param = self.parameter_converter(location, node_schema)
                    if self.ref:
                        param = self._ref(param)
                    params.append(param)

        return params

    def from_path(self, path):
        """
        Create a list of Swagger path params from a cornice service path.

        :type path: string
        :rtype: list
        """
        path_components = path.split("/")
        param_names = [
            comp[1:-1] for comp in path_components if comp.startswith("{") and comp.endswith("}")
        ]

        params = []
        for name in param_names:
            param_schema = colander.SchemaNode(colander.String(), name=name)
            param = self.parameter_converter("path", param_schema)
            if self.ref:
                param = self._ref(param)
            params.append(param)

        return params

    def _ref(self, param, base_name=None):
        """
        Store a parameter schema and return a reference to it.

        :param schema:
            Swagger parameter definition.
        :param base_name:
            Name that should be used for the reference.

        :rtype: dict
        :returns: JSON pointer to the original parameter definition.
        """

        name = base_name or param.get("title", "") or param.get("name", "")

        pointer = self.json_pointer + name
        self.parameter_registry[name] = param

        return {"$ref": pointer}


class ResponseHandler(object):
    """Handles swagger response definitions."""

    json_pointer = "#/responses/"

    def __init__(
        self, definition_handler=DefinitionHandler(), type_converter=TypeConverter(), ref=False
    ):
        """
        :param definition_handler:
            Callable that handles swagger definition schemas.
        :param ref:
            Specifies the ref value when calling from_xxx methods.
        """

        self.response_registry = {}

        self.type_converter = type_converter
        self.definitions = definition_handler
        self.ref = ref

    def from_schema_mapping(self, schema_mapping):
        """
        Creates a Swagger response object from a dict of response schemas.

        :param schema_mapping:
            Dict with entries matching ``{status_code: response_schema}``.
        :rtype: dict
        :returns: Response schema.
        """

        responses = {}

        for status, response_schema in schema_mapping.items():
            response = {}
            if response_schema.description:
                response["description"] = response_schema.description
            else:
                raise CorniceSwaggerException("Responses must have a description.")

            for field_schema in response_schema.children:
                location = field_schema.name

                if location == "body":
                    title = field_schema.__class__.__name__
                    if title == "body":
                        title = response_schema.__class__.__name__ + "Body"
                    field_schema.title = title
                    response["schema"] = self.definitions.from_schema(field_schema)

                elif location in ("header", "headers"):
                    header_schema = self.type_converter(field_schema)
                    headers = header_schema.get("properties")
                    if headers:
                        # Response headers doesn't accept titles
                        for header in headers.values():
                            header.pop("title")

                        response["headers"] = headers

            pointer = response_schema.__class__.__name__
            if self.ref:
                response = self._ref(response, pointer)
            responses[status] = response

        return responses

    def _ref(self, resp, base_name=None):
        """
        Store a response schema and return a reference to it.

        :param schema:
            Swagger response definition.
        :param base_name:
            Name that should be used for the reference.

        :rtype: dict
        :returns: JSON pointer to the original response definition.
        """

        name = base_name or resp.get("title", "") or resp.get("name", "")

        pointer = self.json_pointer + name
        self.response_registry[name] = resp

        return {"$ref": pointer}


class CorniceSwagger(object):
    """Handles the creation of a swagger document from a cornice application."""

    services = []
    """List of cornice services to document. You may use
    `cornice.service.get_services()` to get it."""

    definitions = DefinitionHandler
    """Default :class:`cornice_swagger.swagger.DefinitionHandler` class to use when
    handling OpenAPI schema definitions from kinto.core.cornice payload schemas."""

    parameters = ParameterHandler
    """Default :class:`cornice_swagger.swagger.ParameterHandler` class to use when
    handling OpenAPI operation parameters from kinto.core.cornice request schemas."""

    responses = ResponseHandler
    """Default :class:`cornice_swagger.swagger.ResponseHandler` class to use when
    handling OpenAPI responses from kinto.core.cornice_swagger defined responses."""

    schema_transformers = [body_schema_transformer]
    """List of request schema transformers that should be applied to a request
    schema to make it comply with a cornice default request schema."""

    type_converter = TypeConverter
    """Default :class:`cornice_swagger.converters.schema.TypeConversionDispatcher`
    class used for converting colander schema Types to Swagger Types."""

    parameter_converter = ParameterConverter
    """Default :class:`cornice_swagger.converters.parameters.ParameterConversionDispatcher`
    class used for converting colander/cornice request schemas to Swagger Parameters."""

    custom_type_converters = {}
    """Mapping for supporting custom types conversion on the default TypeConverter.
    Should map `colander.TypeSchema` to `cornice_swagger.converters.schema.TypeConverter`
    callables."""

    default_type_converter = None
    """Supplies a default type converter matching the interface of
    `cornice_swagger.converters.schema.TypeConverter` to be used with unknown types."""

    default_tags = None
    """Provide a default list of tags or a callable that takes a cornice
    service and the method name (e.g GET) and returns a list of Swagger
    tags to be used if not provided by the view."""

    default_op_ids = None
    """Provide a callable that takes a cornice service and the method name
    (e.g. GET) and returns an operation Id that is used if an operation Id is
    not provided. Each operation Id should be unique."""

    default_security = None
    """Provide a default list or a callable that takes a cornice service and
    the method name (e.g. GET) and returns a list of OpenAPI security policies."""

    summary_docstrings = False
    """Enable extracting operation summaries from view docstrings."""

    ignore_methods = ["HEAD", "OPTIONS"]
    """List of service methods that should NOT be presented on the
    documentation. You may use this to remove methods that are not
    essential on the API documentation. Default ignores HEAD and OPTIONS."""

    ignore_ctypes = []
    """List of service content-types that should NOT be presented on the
    documentation. You may use this when a Cornice service definition has
    multiple view definitions for a same method, which is not supported on
    OpenAPI 2.0."""

    api_title = ""
    """Title of the OpenAPI document."""

    api_version = ""
    """Version of the OpenAPI document."""

    base_path = "/"
    """Base path of the documented API. Default is "/"."""

    swagger = {"info": {}}
    """Base OpenAPI document that should be merged with the extracted info
    from the generate call."""

    def __init__(
        self,
        services=None,
        def_ref_depth=0,
        param_ref=False,
        resp_ref=False,
        pyramid_registry=None,
    ):
        """
        :param services:
            List of cornice services to document. You may use
            cornice.service.get_services() to get it.
        :param def_ref_depth:
            How depth swagger object schemas should be split into
            swaggger definitions with JSON pointers. Default (0) is no split.
            You may use negative values to split everything.
        :param param_ref:
            Defines if swagger parameters should be put inline on the operation
            or on the parameters section and referenced by JSON pointers.
            Default is inline.
        :param resp_ref:
            Defines if swagger responses should be put inline on the operation
            or on the responses section and referenced by JSON pointers.
            Default is inline.
        :param pyramid_registry:
            Pyramid registry, should be passed if you use pyramid routes
            instead of service level paths.
        """
        super(CorniceSwagger, self).__init__()

        type_converter = self.type_converter(
            self.custom_type_converters, self.default_type_converter
        )
        parameter_converter = self.parameter_converter(type_converter)
        self.pyramid_registry = pyramid_registry
        if services is not None:
            self.services = services

        # Instantiate handlers
        self.definitions = self.definitions(ref=def_ref_depth, type_converter=type_converter)
        self.parameters = self.parameters(
            self.definitions,
            ref=param_ref,
            type_converter=type_converter,
            parameter_converter=parameter_converter,
        )
        self.responses = self.responses(
            self.definitions, ref=resp_ref, type_converter=type_converter
        )

    def generate(
        self, title=None, version=None, base_path=None, info=None, swagger=None, **kwargs
    ):
        """Generate a Swagger 2.0 documentation. Keyword arguments may be used
        to provide additional information to build methods as such ignores.

        :param title:
            The name presented on the swagger document.
        :param version:
            The version of the API presented on the swagger document.
        :param base_path:
            The path that all requests to the API must refer to.
        :param info:
            Swagger info field.
        :param swagger:
            Extra fields that should be provided on the swagger documentation.

        :rtype: dict
        :returns: Full OpenAPI/Swagger compliant specification for the application.
        """
        title = title or self.api_title
        version = version or self.api_version
        info = info or self.swagger.get("info", {})
        swagger = swagger or self.swagger
        base_path = base_path or self.base_path

        swagger = swagger.copy()
        info.update(title=title, version=version)
        swagger.update(swagger="2.0", info=info, basePath=base_path)

        paths, tags = self._build_paths()

        # Update the provided tags with the extracted ones preserving order
        if tags:
            swagger.setdefault("tags", [])
            tag_names = {t["name"] for t in swagger["tags"]}
            for tag in tags:
                if tag["name"] not in tag_names:
                    swagger["tags"].append(tag)

        # Create/Update swagger sections with extracted values where not provided
        if paths:
            swagger.setdefault("paths", {})
            merge_dicts(swagger["paths"], paths)

        definitions = self.definitions.definition_registry
        if definitions:
            swagger.setdefault("definitions", {})
            merge_dicts(swagger["definitions"], definitions)

        parameters = self.parameters.parameter_registry
        if parameters:
            swagger.setdefault("parameters", {})
            merge_dicts(swagger["parameters"], parameters)

        responses = self.responses.response_registry
        if responses:
            swagger.setdefault("responses", {})
            merge_dicts(swagger["responses"], responses)

        return swagger

    def __call__(self, *args, **kwargs):
        """Deprecated alias of `generate`."""
        self.__dict__.update(**kwargs)

        message = "Calling `CorniceSwagger is deprecated, call `generate` instead"
        warnings.warn(message, DeprecationWarning)
        return self.generate(*args, **kwargs)

    def _check_tags(self, tags):
        """Check if tags was correctly defined as a list"""
        if not isinstance(tags, list):
            raise CorniceSwaggerException("tags should be a list or callable")

    def _get_tags(self, current_tags, new_tags):
        tags = list(current_tags)
        for tag in new_tags:
            root_tag = {"name": tag}
            if root_tag not in tags:
                tags.append(root_tag)
        return tags

    def _build_paths(self):
        """
        Build the Swagger "paths" and "tags" attributes from kinto.core.cornice service
        definitions.
        """
        paths = {}
        tags = []

        for service in self.services:
            path, path_obj = self._extract_path_from_service(service)

            service_tags = getattr(service, "tags", [])
            self._check_tags(service_tags)
            tags = self._get_tags(tags, service_tags)

            for method, view, args in service.definitions:
                if method.lower() in map(str.lower, self.ignore_methods):
                    continue

                op = self._extract_operation_from_view(view, args)

                if any(ctype in op.get("consumes", []) for ctype in self.ignore_ctypes):
                    continue

                # XXX: Swagger doesn't support different schemas for for a same method
                # with different ctypes as cornice. If this happens, you may ignore one
                # content-type from the documentation otherwise we raise an Exception
                # Related to https://github.com/OAI/OpenAPI-Specification/issues/146
                previous_definition = path_obj.get(method.lower())
                if previous_definition:
                    raise CorniceSwaggerException(
                        (
                            "Swagger doesn't support multiple "
                            "views for a same method. You may "
                            "ignore one."
                        )
                    )

                # If tag not defined and a default tag is provided
                if "tags" not in op and self.default_tags:
                    if callable(self.default_tags):
                        op["tags"] = self.default_tags(service, method)
                    else:
                        op["tags"] = self.default_tags

                op_tags = op.get("tags", [])
                self._check_tags(op_tags)

                # Add service tags
                if service_tags:
                    new_tags = service_tags + op_tags
                    op["tags"] = list(OrderedDict.fromkeys(new_tags))

                # Add method tags to root tags
                tags = self._get_tags(tags, op_tags)

                # If operation id is not defined and a default generator is provided
                if "operationId" not in op and self.default_op_ids:
                    if not callable(self.default_op_ids):
                        raise CorniceSwaggerException("default_op_id should be a callable.")
                    op["operationId"] = self.default_op_ids(service, method)

                # If security options not defined and default is provided
                if "security" not in op and self.default_security:
                    if callable(self.default_security):
                        op["security"] = self.default_security(service, method)
                    else:
                        op["security"] = self.default_security

                if not isinstance(op.get("security", []), list):
                    raise CorniceSwaggerException("security should be a list or callable")

                path_obj[method.lower()] = op
            paths[path] = path_obj

        return paths, tags

    def _extract_path_from_service(self, service):
        """
        Extract path object and its parameters from service definitions.

        :param service:
            Cornice service to extract information from.

        :rtype: dict
        :returns: Path definition.
        """

        path_obj = {}
        path = service.path
        route_name = getattr(service, "pyramid_route", None)
        # handle services that don't create fresh routes,
        # we still need the paths so we need to grab pyramid introspector to
        # extract that information
        if route_name:
            # avoid failure if someone forgets to pass registry
            registry = self.pyramid_registry or get_current_registry()
            route_intr = registry.introspector.get("routes", route_name)
            if route_intr:
                path = route_intr["pattern"]
            else:
                msg = "Route `{}` is not found by pyramid introspector".format(route_name)
                raise ValueError(msg)

        # handle traverse and subpath as regular parameters
        # docs.pylonsproject.org/projects/pyramid/en/latest/narr/hybrid.html
        for subpath_marker in ("*subpath", "*traverse"):
            path = path.replace(subpath_marker, "{subpath}")

        # Extract path parameters
        parameters = self.parameters.from_path(path)
        if parameters:
            path_obj["parameters"] = parameters

        return path, path_obj

    def _extract_operation_from_view(self, view, args):
        """
        Extract swagger operation details from colander view definitions.

        :param view:
            View to extract information from.
        :param args:
            Arguments from the view decorator.

        :rtype: dict
        :returns: Operation definition.
        """

        op = {
            "responses": {"default": {"description": "UNDOCUMENTED RESPONSE"}},
        }

        # If 'produces' are not defined in the view, try get from renderers
        renderer = args.get("renderer", "")

        is_json_renderer = (
            "json" in renderer  # allows for "json" or "simplejson"
            or renderer == Service.renderer  # default renderer is json.
        )

        if is_json_renderer:
            produces = ["application/json"]
        elif renderer == "xml":
            produces = ["text/xml"]
        else:
            produces = None

        if produces:
            op.setdefault("produces", produces)

        # Get explicit accepted content-types
        consumes = args.get("content_type")

        if consumes is not None:
            # convert to a list, if it's not yet one
            consumes = to_list(consumes)

            # It is possible to add callables for content_type, so we have to
            # to filter those out, since we cannot evaluate those here.
            consumes = [x for x in consumes if not callable(x)]
            op["consumes"] = consumes

        # Get parameters from view schema
        is_colander = self._is_colander_schema(args)
        if is_colander:
            schema = self._extract_transform_colander_schema(args)
            parameters = self.parameters.from_schema(schema)
        else:
            # Bail out for now
            parameters = None
        if parameters:
            op["parameters"] = parameters

        # Get summary from docstring
        if isinstance(view, str):
            if "klass" in args:
                ob = args["klass"]
                view_ = getattr(ob, view.lower())
                docstring = trim(view_.__doc__)
        else:
            docstring = str(trim(view.__doc__))

        if docstring and self.summary_docstrings:
            op["summary"] = docstring

        # Get response definitions
        if "response_schemas" in args:
            op["responses"] = self.responses.from_schema_mapping(args["response_schemas"])

        # Get response tags
        if "tags" in args:
            op["tags"] = args["tags"]

        # Get response operationId
        if "operation_id" in args:
            op["operationId"] = args["operation_id"]

        # Get security policies
        if "api_security" in args:
            op["security"] = args["api_security"]

        return op

    def _is_colander_schema(self, args):
        schema = args.get("schema")
        return isinstance(schema, colander.Schema) or (
            inspect.isclass(schema) and issubclass(schema, colander.MappingSchema)
        )

    def _extract_transform_colander_schema(self, args):
        """
        Extract schema from view args and transform it using
        the pipeline of schema transformers

        :param args:
            Arguments from the view decorator.

        :rtype: colander.MappingSchema()
        :returns: View schema cloned and transformed
        """

        schema = args.get("schema", colander.MappingSchema())
        if not isinstance(schema, colander.Schema):
            schema = schema()
        schema = schema.clone()
        for transformer in self.schema_transformers:
            schema = transformer(schema, args)
        return schema
